/****************************************************************************
*
*                   SciTech Multi-platform Graphics Library
*
*  ========================================================================
*
*    The contents of this file are subject to the SciTech MGL Public
*    License Version 1.0 (the "License"); you may not use this file
*    except in compliance with the License. You may obtain a copy of
*    the License at http://www.scitechsoft.com/mgl-license.txt
*
*    Software distributed under the License is distributed on an
*    "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
*    implied. See the License for the specific language governing
*    rights and limitations under the License.
*
*    The Original Code is Copyright (C) 1991-1998 SciTech Software, Inc.
*
*    The Initial Developer of the Original Code is SciTech Software, Inc.
*    All Rights Reserved.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Framebuffer access control functions for the MGL.
*
****************************************************************************/

#include "mgl.h"

/*------------------------- Implementation --------------------------------*/

/****************************************************************************
DESCRIPTION:
Enables direct framebuffer access.

HEADER:
mgraph.h

REMARKS:
Enables direct framebuffer access so that you can directly rasterize to the
linear framebuffer memory using your own custom routines. Note that calling
this function is absolutely necessary when using hardware acceleration, as
this function correctly arbitrates between the hardware accelerator graphics
engine and your direct framebuffer rasterizing code.

SEE ALSO:
MGL_endDirectAccess, MGL_endDirectAccessExt
****************************************************************************/
void MGLAPI MGL_beginDirectAccess(void)
{
    if (DC.r.BeginDirectAccess)
        DC.r.BeginDirectAccess(&DC);
}

/****************************************************************************
DESCRIPTION:
Disables direct framebuffer access.

HEADER:
mgraph.h

REMARKS:
Disables direct framebuffer access so that you can use the accelerator
functions to draw to the framebuffer memory. Note that calling this function
is absolutely necessary when using hardware acceleration, as this function
and the corresponding MGL_beginDirectAccess correctly arbitrate between the
hardware accelerator graphics engine and your direct framebuffer writes.

SEE ALSO:
MGL_beginDirectAccess, MGL_endDirectAccessExt
****************************************************************************/
void MGLAPI MGL_endDirectAccess(void)
{
    if (DC.r.EndDirectAccess)
        DC.r.EndDirectAccess(&DC);
}

/****************************************************************************
DESCRIPTION:
Disables direct framebuffer access with shadow buffer update.

HEADER:
mgraph.h

REMARKS:
This function is used to disable direct framebuffer access just like the
regular MGL_endDirectAccess function. However it is designed to be more
efficient for application that are running with a system memory shadow
shadow buffer. The MGL will be using a system memory shadow buffer if
the underlying device driver does not provide a hardware linear framebuffer
(ie: a Standard VGA device or a VESA VBE 1.2 device). In the case where a
shadow buffer is in use, only the region specified by the coordinates will
be updated.

Internally all MGL functions automatically handle shadow buffer updates
without any application programmer knowledge, so this is only needed if your
code is doing any custom rendering that must work correctly on devices that
use a shadow buffer.

SEE ALSO:
MGL_beginDirectAccess, MGL_endDirectAccess
****************************************************************************/
void MGLAPI MGL_endDirectAccessExt(
    int left,
    int top,
    int right,
    int bottom)
{
    if (DC.r.EndDirectAccessExt)
        DC.r.EndDirectAccessExt(&DC,left,top,right,bottom);
    else if (DC.r.EndDirectAccess)
        DC.r.EndDirectAccess(&DC);
}

/****************************************************************************
DESCRIPTION:
Setup for high speed pixel drawing.

HEADER:
mgraph.h

REMARKS:
Sets up the video hardware for plotting single pixels as fast a possible. You
must call this routine before calling any of the MGL_pixel and MGL_getPixel
routines to ensure correct operation, and you must call the MGL_endPixel
routine after you have finished.

This routine is intended primarily to ensure fast operation if you need to
plot more than a single pixel at a time.

SEE ALSO:
MGL_endPixel, MGL_pixel, MGL_getPixel.
****************************************************************************/
void MGLAPI MGL_beginPixel(void)
{
    UPDATE_VISIBLE_CLIP_LIST(&DC);
    if (DC.r.BeginDirectAccess)
        DC.r.BeginDirectAccess(&DC);
}

/****************************************************************************
DESCRIPTION:
Ends high speed pixel drawing operation.

HEADER:
mgraph.h

REMARKS:
This function ends a set of high speed pixel drawing operations, started
with a call to MGL_beginPixel. This routine is intended primarily to ensure
fast operation if you intend to plot more than a single pixel at a time.

SEE ALSO:
MGL_beginPixel
****************************************************************************/
void MGLAPI MGL_endPixel(void)
{
    /* Pixels are handled directly in the shadow buffer code */
    if (DC.r.EndDirectAccess && !(DC.flags & MGL_SHADOW_BUFFER))
        DC.r.EndDirectAccess(&DC);
}

